"""Weight 3 harmonic sums."""

import numba as nb

from eko.constants import log2, zeta2, zeta3

from . import g_functions as gf
from .polygamma import cern_polygamma, symmetry_factor


@nb.njit(cache=True)
def S3(N):
    r"""Compute the harmonic sum :math:`S_3(N)`.

    .. math::
      S_3(N) = \sum\limits_{j=1}^N \frac 1 {j^3} = \frac 1 2 \psi_2(N+1)+\zeta(3)

    with :math:`\psi_2(N)` the 2nd-polygamma function and :math:`\zeta` the
    Riemann zeta function.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    S_3 : complex
        Harmonic sum :math:`S_3(N)`

    See Also
    --------
    ekore.harmonics.polygamma.cern_polygamma : :math:`\psi_k(N)`

    """
    return 0.5 * cern_polygamma(N + 1.0, 2) + zeta3


@nb.njit(cache=True)
def Sm3(N, hS3, hS3mh, hS3h, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-3}(N)`.

    .. math::
      S_{-3}(N) = \sum\limits_{j=1}^N \frac {(-1)^j} {j^3}

    Parameters
    ----------
    N : complex
        Mellin moment
    hS3:  complex
        Harmonic sum :math:`S_{3}(N)`
    hS3mh: complex
        Harmonic sum :math:`S_{3}((N-1)/2)`
    hS3h: complex
        Harmonic sum :math:`S_{3}(N/2)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N = 1`),
        False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm3 : complex
        Harmonic sum :math:`S_{-3}(N)`

    See Also
    --------
    ekore.harmonics.w3.S3 : :math:`S_3(N)`

    """
    if is_singlet is None:
        return (
            1 / 2**2 * ((1 - (-1) ** N) / 2 * hS3mh + ((-1) ** N + 1) / 2 * hS3h)
            - hS3
        )
    if is_singlet:
        return 1 / 2**2 * hS3h - hS3
    return 1 / 2**2 * hS3mh - hS3


@nb.njit(cache=True)
def S21(N, S1, S2):
    r"""Analytic continuation of harmonic sum :math:`S_{2,1}(N)`.

    As implemented in :eqref:`B.5.77` of :cite:`MuselliPhD` and :eqref:`37` of
    :cite:`Bl_mlein_2000`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`

    Returns
    -------
    S21 : complex
        Harmonic sum :math:`S_{2,1}(N)`

    """
    return -gf.mellin_g18(N, S1, S2) + 2 * zeta3


@nb.njit(cache=True)
def Sm21(N, S1, Sm1, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-2,1}(N)`.

    As implemented in :eqref:`B.5.75` of :cite:`MuselliPhD` and :eqref:`22` of
    :cite:`Bl_mlein_2000`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1:  complex
        Harmonic sum :math:`S_{1}(N)`
    Sm1: complex
        Harmonic sum :math:`S_{-1}(N)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N = 1`),
        False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm21 : complex
        Harmonic sum :math:`S_{-2,1}(N)`

    """
    # Note mellin g3 was integrated following x^(N-1) convention.
    eta = symmetry_factor(N, is_singlet)
    return (
        -eta * gf.mellin_g3(N + 1, S1 + 1 / (N + 1))
        + zeta2 * Sm1
        - 5 / 8 * zeta3
        + zeta2 * log2
    )


@nb.njit(cache=True)
def S2m1(N, S2, Sm1, Sm2, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{2,-1}(N)`.

    As implemented in :eqref:`B.5.76` of :cite:`MuselliPhD` and :eqref:`23` of
    :cite:`Bl_mlein_2000`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    Sm1: complex
        Harmonic sum :math:`S_{-1}(N)`
    Sm2: complex
        Harmonic sum :math:`S_{-2}(N)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N =
        1`), False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    S2m1 : complex
        Harmonic sum :math:`S_{2,-1}(N)`

    """
    eta = symmetry_factor(N, is_singlet)
    return (
        -eta * gf.mellin_g4(N)
        - log2 * (S2 - Sm2)
        - 1 / 2 * zeta2 * Sm1
        + 1 / 4 * zeta3
        - 1 / 2 * zeta2 * log2
    )


@nb.njit(cache=True)
def Sm2m1(N, S1, S2, Sm2):
    r"""Analytic continuation of harmonic sum :math:`S_{-2,-1}(N)`.

    As implemented in :eqref:`B.5.74` of :cite:`MuselliPhD` and :eqref:`38` of
    :cite:`Bl_mlein_2000`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    Sm2: complex
        Harmonic sum :math:`S_{-2}(N)`

    Returns
    -------
    Sm2m1 : complex
        Harmonic sum :math:`S_{-2,-1}(N)`

    """
    return -gf.mellin_g19(N, S1) + log2 * (S2 - Sm2) - 5 / 8 * zeta3



r"""Implementation of Mellin transformation of logarithms.

We provide transforms of:

- :math:`(1-x)\ln^k(1-x), \quad k = 1,2,3`
- :math:`\ln^k(1-x), \quad k = 1,3,4,5`

"""
import numba as nb


@nb.njit(cache=True)
def lm11m1(n, S1):
    r"""Mellin transform of :math:`(1-x)\ln(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)\ln(1-x)](N)`
    """
    return 1 / (1 + n) ** 2 - S1 / (1 + n) ** 2 - S1 / (n * (1 + n) ** 2)


@nb.njit(cache=True)
def lm12m1(n, S1, S2):
    r"""Mellin transform of :math:`(1-x)\ln^2(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)\ln^2(1-x)](N)`

    """
    return (
        -2 / (1 + n) ** 3
        - (2 * S1) / (1 + n) ** 2
        + S1**2 / n
        - S1**2 / (1 + n)
        + S2 / n
        - S2 / (1 + n)
    )


@nb.njit(cache=True)
def lm13m1(n, S1, S2, S3):
    r"""Mellin transform of :math:`(1-x)\ln^3(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)\ln^3(1-x)](N)`

    """
    return (
        3 * n * (1 + n) ** 2 * S1**2
        - (1 + n) ** 3 * S1**3
        + 3 * n * (1 + n) ** 2 * S2
        - 3 * (1 + n) * S1 * (-2 * n + (1 + n) ** 2 * S2)
        - 2 * (-3 * n + (1 + n) ** 3 * S3)
    ) / (n * (1 + n) ** 4)


@nb.njit(cache=True)
def lm14m1(n, S1, S2, S3, S4):
    r"""Mellin transform of :math:`(1-x)\ln^4(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`
    S4 : complex
        Harmonic sum :math:`S_{4}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)\ln^4(1-x)](N)`

    """
    return (
        -24 * n
        - 4 * n * (1 + n) ** 3 * S1**3
        + (1 + n) ** 4 * S1**4
        - 12 * n * (1 + n) ** 2 * S2
        + 3 * (1 + n) ** 4 * S2**2
        + 6 * (1 + n) ** 2 * S1**2 * (-2 * n + (1 + n) ** 2 * S2)
        - 8 * n * S3
        - 24 * n**2 * S3
        - 24 * n**3 * S3
        - 8 * n**4 * S3
        - 4
        * (1 + n)
        * S1
        * (3 * n * (1 + n) ** 2 * S2 - 2 * (-3 * n + (1 + n) ** 3 * S3))
        + 6 * S4
        + 24 * n * S4
        + 36 * n**2 * S4
        + 24 * n**3 * S4
        + 6 * n**4 * S4
    ) / (n * (1 + n) ** 5)


@nb.njit(cache=True)
def lm11(n, S1):
    r"""Mellin transform of :math:`\ln(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[\ln(1-x)](N)`

    """
    return -S1 / n


@nb.njit(cache=True)
def lm12(n, S1, S2):
    r"""Mellin transform of :math:`\ln^2(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[\ln^2(1-x)](N)`

    """
    return (S1**2 + S2) / n


@nb.njit(cache=True)
def lm13(n, S1, S2, S3):
    r"""Mellin transform of :math:`\ln^3(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[\ln^3(1-x)](N)`

    """
    return -((S1**3 + 3 * S1 * S2 + 2 * S3) / n)


@nb.njit(cache=True)
def lm14(n, S1, S2, S3, S4):
    r"""Mellin transform of :math:`\ln^4(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`
    S4 : complex
        Harmonic sum :math:`S_{4}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[\ln^4(1-x)](N)`

    """
    return (S1**4 + 6 * S1**2 * S2 + 3 * S2**2 + 8 * S1 * S3 + 6 * S4) / n


@nb.njit(cache=True)
def lm15(n, S1, S2, S3, S4, S5):
    r"""Mellin transform of :math:`\ln^5(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`
    S4 : complex
        Harmonic sum :math:`S_{4}(N)`
    S5 : complex
        Harmonic sum :math:`S_{5}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[\ln^5(1-x)](N)`

    """
    return (
        -(
            S1**5
            + 10 * S1**3 * S2
            + 20 * S1**2 * S3
            + 15 * S1 * (S2**2 + 2 * S4)
            + 4 * (5 * S2 * S3 + 6 * S5)
        )
        / n
    )


@nb.njit(cache=True)
def lm11m2(n, S1):
    r"""Mellin transform of :math:`(1-x)^2\ln(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)^2\ln(1-x)](N)`

    """
    return (5 + 3 * n - (2 * (1 + n) * (2 + n) * S1) / n) / (
        (1 + n) ** 2 * (2 + n) ** 2
    )


@nb.njit(cache=True)
def lm12m2(n, S1, S2):
    r"""Mellin transform of :math:`(1-x)^2\ln^2(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)^2\ln^2(1-x)](N)`

    """
    return (
        2
        * (
            n * (-9 - 8 * n + n**3)
            - n * (10 + 21 * n + 14 * n**2 + 3 * n**3) * S1
            + (2 + 3 * n + n**2) ** 2 * S1**2
            + (2 + 3 * n + n**2) ** 2 * S2
        )
    ) / (n * (1 + n) ** 3 * (2 + n) ** 3)


@nb.njit(cache=True)
def lm13m2(n, S1, S2, S3):
    r"""Mellin transform of :math:`(1-x)^2\ln^3(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)^2\ln^3(1-x)](N)`

    """
    return (
        -6 * n * (-17 - 21 * n - 2 * n**2 + 6 * n**3 + 2 * n**4)
        + 3 * n * (5 + 3 * n) * (2 + 3 * n + n**2) ** 2 * S1**2
        - 2 * (2 + 3 * n + n**2) ** 3 * S1**3
        + 3 * n * (5 + 3 * n) * (2 + 3 * n + n**2) ** 2 * S2
        - 6
        * (2 + 3 * n + n**2)
        * S1
        * (n * (-9 - 8 * n + n**3) + (2 + 3 * n + n**2) ** 2 * S2)
        - 4 * (2 + 3 * n + n**2) ** 3 * S3
    ) / (n * (1 + n) ** 4 * (2 + n) ** 4)


@nb.njit(cache=True)
def lm14m2(n, S1, S2, S3, S4):
    r"""Mellin transform of :math:`(1-x)^2\ln^4(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`
    S4 : complex
        Harmonic sum :math:`S_{4}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)^2\ln^4(1-x)](N)`

    """
    return (
        2
        / (n * (1 + n) ** 5 * (2 + n) ** 5)
        * (
            +12 * n * (-33 + n * (-54 + n * (-15 + n * (20 + 3 * n * (5 + n)))))
            - 2 * n * (1 + n) ** 3 * (2 + n) ** 3 * (5 + 3 * n) * S1**3
            + (1 + n) ** 4 * (2 + n) ** 4 * S1**4
            + 6 * n * (1 + n) ** 2 * (2 + n) ** 2 * (-9 - 8 * n + n**3) * S2
            + 3 * (1 + n) ** 4 * (2 + n) ** 4 * S2**2
            + 6
            * (1 + n) ** 2
            * (2 + n) ** 2
            * S1**2
            * (n * (-9 - 8 * n + n**3) + (1 + n) ** 2 * (2 + n) ** 2 * S2)
            - 4 * n * (1 + n) ** 3 * (2 + n) ** 3 * (5 + 3 * n) * S3
            + 2
            * (1 + n)
            * (2 + n)
            * S1
            * (
                6 * n * (-17 + n * (-21 + 2 * n * (-1 + n * (3 + n))))
                - 3 * n * (1 + n) ** 2 * (2 + n) ** 2 * (5 + 3 * n) * S2
                + 4 * (1 + n) ** 3 * (2 + n) ** 3 * S3
            )
            + 6 * (1 + n) ** 4 * (2 + n) ** 4 * S4
        )
    )



"""Define output representation structures."""
import contextlib
import copy
import logging
import os
import pathlib
import shutil
import tarfile
import tempfile
from dataclasses import dataclass
from typing import List, Optional

import numpy as np
import yaml

from .. import interpolation
from . import exceptions, raw
from .access import AccessConfigs
from .bases import Bases
from .inventory import Inventory
from .items import Evolution, Matching, Operator, Recipe, Target
from .metadata import Metadata
from .paths import InternalPaths
from .runcards import OperatorCard, TheoryCard
from .types import EvolutionPoint as EPoint
from .types import SquaredScale

logger = logging.getLogger(__name__)

TEMP_PREFIX = "eko-"


def inventories(path: pathlib.Path, access: AccessConfigs) -> dict:
    """Set up empty inventories for object initialization."""
    paths = InternalPaths(path)
    return dict(
        recipes=Inventory(
            paths.recipes, access, Evolution, contentless=True, name="recipes"
        ),
        recipes_matching=Inventory(
            paths.recipes_matching,
            access,
            Matching,
            contentless=True,
            name="matching-recipes",
        ),
        parts=Inventory(paths.parts, access, Evolution, name="parts"),
        parts_matching=Inventory(
            paths.parts_matching, access, Matching, name="matching-parts"
        ),
        operators=Inventory(paths.operators, access, Target, name="operators"),
    )


@dataclass
class EKO:
    """Operator interface.

    This class offers an interface to an abstract operator, between memory and
    disk.

    An actual operator might be arbitrarily huge, and in particular size
    limitations in memory are far more strict than on disk.
    Since manually managing, for each application, the burden of off-loading
    part of the operator might be hard and occasionally not possible (without a
    clear picture of the internals), the library itself offers this facility.

    In particular, the data format on disk has a complete specification, and
    can hold a full operator independently of the loading procedure.
    In order to accomplish the former goal, the remaining task of partial
    loading is done by this class (for the Python library, other
    implementations are possible and encouraged).

    For this reason, a core component of an :class:`EKO` object is a path,
    referring to the location on disk of the corresponding operator.
    Any :class:`EKO` has an associated path:

        - for the computed object, it corresponds to the path where the actual
          result of the computation is already saved
        - for a new object, it is the path at which any result of final or
          intermediate computation is stored, as soon as it is produced

    The computation can be stopped at any time, without the loss of any of the
    intermediate results.

    """

    recipes: Inventory[Evolution]
    recipes_matching: Inventory[Matching]
    parts: Inventory[Evolution]
    parts_matching: Inventory[Matching]
    operators: Inventory[Target]

    # public containers
    # -----------------
    metadata: Metadata
    """Operator metadata."""
    access: AccessConfigs
    """Access related configurations."""

    # shortcut properties
    # -------------------
    @property
    def paths(self) -> InternalPaths:
        """Accessor for internal paths."""
        return InternalPaths(self.metadata.path)

    @property
    def bases(self) -> Bases:
        """Bases information."""
        return self.metadata.bases

    @property
    def xgrid(self) -> interpolation.XGrid:
        """Momentum fraction internal grid."""
        return self.bases.xgrid

    @xgrid.setter
    def xgrid(self, value: interpolation.XGrid):
        """Set `xgrid` value."""
        self.bases.xgrid = value
        self.update()

    @property
    def mu20(self) -> SquaredScale:
        """Provide squared initial scale."""
        return self.metadata.origin[0]

    @property
    def mu2grid(self) -> List[SquaredScale]:
        """Provide the list of :math:`Q^2` as an array."""
        return [ep.scale for ep in self.operators]

    @property
    def evolgrid(self) -> List[EPoint]:
        """Provide the list of evolution points as an array."""
        return list(self)

    @property
    def theory_card(self):
        """Provide theory card, retrieving from the dump."""
        return TheoryCard.from_dict(
            yaml.safe_load(self.paths.theory_card.read_text(encoding="utf-8"))
        )

    @property
    def operator_card(self):
        """Provide operator card, retrieving from the dump."""
        return OperatorCard.from_dict(
            yaml.safe_load(self.paths.operator_card.read_text(encoding="utf-8"))
        )

    # persistency control
    # -------------------

    def update(self):
        """Write updates to structure for persistency."""
        self.access.assert_writeable()
        self.metadata.update()

    def assert_permissions(self, read=True, write=False):
        """Assert permissions on current operator."""
        if read:
            self.access.assert_open()
        if write:
            self.access.assert_writeable()

    @property
    def permissions(self):
        """Provide permissions information."""
        return dict(read=self.access.read, write=self.access.write)

    # recipes management
    # -------------------

    def load_recipes(self, recipes: List[Recipe]):
        """Load recipes in bulk."""
        for recipe in recipes:
            # leverage auto-save
            if isinstance(recipe, Evolution):
                self.recipes[recipe] = None
            else:
                self.recipes_matching[recipe] = None

    # operator management
    # -------------------
    def __getitem__(self, ep: EPoint) -> Optional[Operator]:
        r"""Retrieve operator for given evolution point."""
        return self.operators[Target.from_ep(ep)]

    def __setitem__(self, ep: EPoint, op: Operator):
        """Set operator associated to an evolution point."""
        self.operators[Target.from_ep(ep)] = op

    def __delitem__(self, ep: EPoint):
        """Drop operator from memory."""
        del self.operators[Target.from_ep(ep)]

    def __delattr__(self, name: str):
        """Empty an inventory cache."""
        attr = getattr(self, name)
        if isinstance(attr, Inventory):
            attr.empty()
        else:
            super().__delattr__(name)

    @contextlib.contextmanager
    def operator(self, ep: EPoint):
        """Retrieve an operator and discard it afterwards.

        To be used as a contextmanager: the operator is automatically loaded as
        usual, but on the closing of the context manager it is dropped from
        memory.

        """
        try:
            yield self[ep]
        finally:
            del self[ep]

    def __iter__(self):
        """Iterate over keys (i.e. evolution points)."""
        for target in self.operators:
            yield target.ep

    def items(self):
        """Iterate operators, with minimal load.

        Pay attention, this iterator:

        - is not a read-only operation from the point of view of the in-memory
          object (since the final result after iteration is no operator loaded)
        - but it is a read-only operation from the point of view of the
          permanent object on-disk

        Yields
        ------
        tuple
            couples of ``(q2, operator)``, loaded immediately before, unloaded
            immediately after

        """
        for target in self.operators:
            # recast to evolution point
            ep = target.ep

            # auto-load
            op = self[ep]
            assert op is not None
            yield ep, op
            # auto-unload
            del self[ep]

    def __contains__(self, ep: EPoint) -> bool:
        """Check whether the operator related to the evolution point is present.

        'Present' means, in this case, they are available in the :class:`EKO`.
        But it is telling nothing about being loaded in memory or not.

        """
        return Target.from_ep(ep) in self.operators

    def approx(
        self, ep: EPoint, rtol: float = 1e-6, atol: float = 1e-10
    ) -> Optional[EPoint]:
        r"""Look for close enough evolution point in the :class:`EKO`.

        The distance is mostly evaluated along the :math:`\mu^2` dimension,
        while :math:`n_f` is considered with a discrete distance: if two points
        have not the same, they are classified as far.

        Raises
        ------
        ValueError
            if multiple values are find in the neighbourhood

        """
        eps = np.array([ep_ for ep_ in self if ep_[1] == ep[1]])
        mu2s = np.array([mu2 for mu2, _ in eps])
        close = eps[np.isclose(ep[0], mu2s, rtol=rtol, atol=atol)]

        if len(close) == 1:
            return tuple(close[0])
        if len(close) == 0:
            return None
        raise ValueError(f"Multiple values of Q2 have been found close to {ep}")

    def unload(self):
        """Fully unload the operators in memory."""
        for ep in self:
            del self[ep]

    # operator management
    # -------------------

    def deepcopy(self, path: os.PathLike):
        """Create a deep copy of current instance.

        The managed on-disk object is copied as well, to the new ``path``
        location.
        If you don't want to copy the disk, consider using directly::

            copy.deepcopy(myeko)

        It will perform the exact same operation, without propagating it to the
        disk counterpart.

        Parameters
        ----------
        path :
            path to the permanent location of the new object (not the temporary
            directory)

        Returns
        -------
        EKO
            the copy created

        """
        # deepcopy is tuned to copy an open operator into a closed one, since
        # operators are always created closed (to avoid delegating the user to
        # close an object he never opened), then let's assert this assumption
        self.access.assert_open()

        # return the object fully on disk, in order to avoid expensive copies
        # also in memory (they would be copied on-disk in any case)
        self.unload()

        new = copy.deepcopy(self)
        new.access.path = pathlib.Path(path)
        new.access.readonly = False
        new.access.open = True

        tmpdir = pathlib.Path(tempfile.mkdtemp(prefix=TEMP_PREFIX))
        new.metadata.path = tmpdir
        # copy old dir to new dir
        tmpdir.rmdir()
        shutil.copytree(self.paths.root, new.paths.root)
        new.close()

    @staticmethod
    def load(tarpath: os.PathLike, dest: os.PathLike):
        """Load the content of archive in a target directory.

        Parameters
        ----------
        tarpath: os.PathLike
            the archive to extract
        tmppath: os.PathLike
            the destination directory

        """
        try:
            with tarfile.open(tarpath) as tar:
                raw.safe_extractall(tar, dest)
        except tarfile.ReadError:
            raise exceptions.OutputNotTar(f"Not a valid tar archive: '{tarpath}'")

    @classmethod
    def open(cls, path: os.PathLike, mode="r"):
        """Open EKO object in the specified mode."""
        path = pathlib.Path(path)
        access = AccessConfigs(path, readonly=False, open=True)
        load = False
        if mode == "r":
            load = True
            access.readonly = True
        elif mode in "w":
            pass
        elif mode in "a":
            load = True
        else:
            raise ValueError(f"Unknown file mode: {mode}")

        tmpdir = pathlib.Path(tempfile.mkdtemp(prefix=TEMP_PREFIX))
        if load:
            cls.load(path, tmpdir)
            metadata = Metadata.load(tmpdir)
            opened = cls(
                **inventories(tmpdir, access),
                metadata=metadata,
                access=access,
            )
            opened.operators.sync()
        else:
            opened = Builder(path=tmpdir, access=access)

        return opened

    @classmethod
    def read(cls, path: os.PathLike):
        """Read the content of an EKO.

        Type-safe alias for::

            EKO.open(... , "r")

        """
        eko = cls.open(path, "r")
        assert isinstance(eko, EKO)
        return eko

    @classmethod
    def create(cls, path: os.PathLike):
        """Create a new EKO.

        Type-safe alias for::

            EKO.open(... , "w")

        """
        builder = cls.open(path, "w")
        assert isinstance(builder, Builder)
        return builder

    @classmethod
    def edit(cls, path: os.PathLike):
        """Read from and write on existing EKO.

        Type-safe alias for::

            EKO.open(... , "a")

        """
        eko = cls.open(path, "a")
        assert isinstance(eko, EKO)
        return eko

    def __enter__(self):
        """Allow EKO to be used in :obj:`with` statements."""
        return self

    def dump(self, archive: Optional[os.PathLike] = None):
        """Dump the current content to archive.

        Parameters
        ----------
        archive: os.PathLike or None
            path to archive, in general you should keep the default, that will
            make use of the registered path (default: ``None``)

        Raises
        ------
        ValueError
            when trying to dump on default archive in read-only mode

        """
        if archive is None:
            self.access.assert_writeable(
                "Not possible to dump on default archive in read-only mode"
            )
            archive = self.access.path

        with tarfile.open(archive, "w") as tar:
            tar.add(self.metadata.path, arcname=".")

    def close(self):
        """Close the current object, cleaning up.

        If not in read-only mode, dump to permanent storage.
        Remove the temporary directory used.

        """
        if not self.access.readonly:
            # clean given path, to overwrite it - default 'w'rite behavior
            self.access.path.unlink(missing_ok=True)
            self.dump()

        self.access.open = False
        shutil.rmtree(self.metadata.path)

    def __exit__(self, exc_type: type, _exc_value, _traceback):
        """Ensure EKO to be closed properly."""
        if exc_type is not None:
            return

        self.close()

    @property
    def raw(self) -> dict:
        """Provide raw representation of the full content.

        Returns
        -------
        dict
            nested dictionary, storing all the values in the structure, but the
            operators themselves

        """
        return dict(mu2grid=self.mu2grid, metadata=self.metadata.raw)


@dataclass
class Builder:
    """Build EKO instances."""

    path: pathlib.Path
    """Path on disk to ."""
    access: AccessConfigs
    """Access related configurations."""

    # optional arguments, required at build time
    theory: Optional[TheoryCard] = None
    operator: Optional[OperatorCard] = None

    eko: Optional[EKO] = None

    def __post_init__(self):
        """Validate paths."""
        if self.access.path.suffix != ".tar":
            raise exceptions.OutputNotTar(self.access.path)
        if self.access.path.exists():
            raise exceptions.OutputExistsError(self.access.path)

    def load_cards(self, theory: TheoryCard, operator: OperatorCard):
        """Load both theory and operator card."""
        self.theory = theory
        self.operator = operator

        return self

    def build(self) -> EKO:
        """Build EKO instance.

        Returns
        -------
        EKO
            the constructed instance

        Raises
        ------
        RuntimeError
            if not enough information is available (at least one card missing)

        """
        missing = []
        for card in ["theory", "operator"]:
            if getattr(self, card) is None:
                missing.append(card)

        if len(missing) > 0:
            raise RuntimeError(
                f"Can not build an EKO, since following cards are missing: {missing}"
            )

        # tell the static analyzer as well
        assert self.theory is not None
        assert self.operator is not None

        self.access.open = True
        metadata = Metadata(
            _path=self.path,
            origin=(self.operator.mu20, self.theory.heavy.num_flavs_init),
            bases=Bases(xgrid=self.operator.xgrid),
        )
        InternalPaths(self.path).bootstrap(
            theory=self.theory.raw,
            operator=self.operator.raw,
            metadata=metadata.raw,
        )

        self.eko = EKO(
            **inventories(self.path, self.access),
            metadata=metadata,
            access=self.access,
        )

        return self.eko

    def __enter__(self):
        """Allow Builder to be used in :obj:`with` statements."""
        return self

    def __exit__(self, exc_type: type, _exc_value, _traceback):
        """Ensure EKO to be closed properly."""
        if exc_type is not None:
            return

        # assign to variable to help type checker, otherwise self.eko might be
        # a property, and its type can change at every evaluation
        eko = self.eko
        if eko is not None:
            eko.close()



"""Module containing the harmonics sums implementation.

Definitions are coming from :cite:`MuselliPhD,Bl_mlein_2000,Blumlein:2009ta`.
"""
from . import cache, g_functions, polygamma
from .w1 import S1, Sm1
from .w2 import S2, Sm2
from .w3 import S3, S21, S2m1, Sm2m1, Sm3, Sm21
from .w4 import S4, S31, S211, Sm4, Sm22, Sm31, Sm211
from .w5 import S5, Sm5



"""Sets the physical constants."""

import numba as nb
import numpy as np
from scipy.special import zeta

NC = 3
"""The number of colors."""

TR = float(1.0 / 2.0)
"""The normalization of fundamental generators.

Defaults to :math:`T_R = 1/2`.
"""

CA = float(NC)
"""Second Casimir constant in the adjoint representation.

Defaults to :math:`N_C = 3`.
"""

CF = float((NC * NC - 1.0) / (2.0 * NC))
r"""Second Casimir constant in the fundamental representation.

Defaults to :math:`\frac{N_C^2-1}{2N_C} = 4/3`.
"""

eu2 = 4.0 / 9
"""Up quarks charge squared."""

ed2 = 1.0 / 9
"""Down quarks charge squared."""


zeta2 = zeta(2)
r""":math:`\zeta(2)`"""

zeta3 = zeta(3)
r""":math:`\zeta(3)`"""

zeta4 = zeta(4)
r""":math:`\zeta(4)`"""

zeta5 = zeta(5)
r""":math:`\zeta(5)`"""

log2 = np.log(2)
r""":math:`\ln(2)`"""

li4half = 0.517479
""":math:`Li_{4}(1/2)`"""


def update_colors(nc):
    """Update the number of colors to :math:`NC = nc`.

    The Casimirs for a generic value of :math:`NC` are consistenly updated as
    well.

    Parameters
    ----------
    nc : int
      Number of colors

    """
    global NC, CA, CF  # pylint: disable=global-statement

    NC = int(nc)
    CA = float(NC)
    CF = float((NC * NC - 1.0) / (2.0 * NC))


@nb.njit(cache=True)
def uplike_flavors(nf):
    """Compute the number of up flavors.

    Parameters
    ----------
    nf : int
        Number of active flavors

    Returns
    -------
    nu : int

    """
    if nf not in range(2, 6 + 1):
        raise NotImplementedError("Selected nf is not implemented")
    nu = nf // 2
    return nu


@nb.njit(cache=True)
def charge_combinations(nf):
    """
    Compute the combination of charges.

    Parameters
    ----------
    nf : int
        Number of active flavors

    Returns
    -------
    e2avg : float
    vue2m : float
    vde2m : float

    """
    nu = uplike_flavors(nf)
    nd = nf - nu
    e2avg = (nu * eu2 + nd * ed2) / nf
    vue2m = nu / nf * (eu2 - ed2)
    vde2m = nd / nf * (eu2 - ed2)
    e2delta = vde2m - vue2m + e2avg
    return e2avg, vue2m, vde2m, e2delta



"""Define paths inside an `eko.EKO` object."""
import pathlib
from dataclasses import dataclass

import yaml

THEORYFILE = "theory.yaml"
OPERATORFILE = "operator.yaml"
METADATAFILE = "metadata.yaml"
RECIPESDIR = "recipes"
PARTSDIR = "parts"
MATCHINGDIR = "matching"
OPERATORSDIR = "operators"


@dataclass
class InternalPaths:
    """Paths inside an EKO folder.

    This structure exists to locate in a single place the internal structure of
    an EKO folder.

    The only value required is the root path, everything else is computed
    relative to this root.
    In case only the relative paths are required, just create this structure
    with :attr:`root` equal to emtpty string or ``"."``.

    """

    root: pathlib.Path
    "The root of the EKO folder (use placeholder if not relevant)"

    @property
    def metadata(self):
        """Metadata file."""
        return self.root / METADATAFILE

    @property
    def recipes(self):
        """Recipes folder."""
        return self.root / RECIPESDIR

    @property
    def recipes_matching(self):
        """Matching recipes folder."""
        return self.root / RECIPESDIR / MATCHINGDIR

    @property
    def parts(self):
        """Parts folder."""
        return self.root / PARTSDIR

    @property
    def parts_matching(self):
        """Matching parts folder."""
        return self.root / PARTSDIR / MATCHINGDIR

    @property
    def operators(self):
        """Operators folder.

        This is the one containing the actual EKO components, after
        computation has been performed.

        """
        return self.root / OPERATORSDIR

    @property
    def theory_card(self):
        """Theory card dump."""
        return self.root / THEORYFILE

    @property
    def operator_card(self):
        """Operator card dump."""
        return self.root / OPERATORFILE

    def bootstrap(self, theory: dict, operator: dict, metadata: dict):
        """Create directory structure."""
        self.metadata.write_text(yaml.dump(metadata), encoding="utf-8")
        self.theory_card.write_text(yaml.dump(theory), encoding="utf-8")
        self.operator_card.write_text(yaml.dump(operator), encoding="utf-8")
        self.recipes.mkdir()
        self.recipes_matching.mkdir()
        self.parts.mkdir()
        self.parts_matching.mkdir()
        self.operators.mkdir()



r"""The |NLO| |OME| in the polarized case for the matching conditions in the |VFNS|.

Heavy quark contribution for intrinsic evolution are not considered for the polarized case.
The matching conditions for the |VFNS| at :math:`\mu_F^2 \neq m_H^2` are provided in :cite:`Bierenbaum:2022biv`.
In the paper, the fraction :math:`\mu_F^2 / m_H^2` inside the log is inverted, yielding an additional factor of (-1) wherever ``L`` has an odd power.
Additionally, a different convention for the anomalous dimensions is used, yielding a factor 2 in the |OME|'s wherever they are present.
The anomalous dimensions and beta function with the addition 'hat' have the form :math:`\hat\gamma = \gamma^{(nf+1)} - \gamma^{(nf)}`.
"""
import numba as nb
import numpy as np

from ....anomalous_dimensions.polarized.space_like.as1 import gamma_qg as gamma0_qg
from ...unpolarized.space_like.as1 import A_gg as A_gg_unpol


@nb.njit(cache=True)
def A_hg(n, L):
    r"""Compute the |NLO| heavy-gluon |OME| :math:`A_{Hg}^{S,(1)}`.

    Implements :eqref:`104` of :cite:`Bierenbaum:2022biv`.

    Parameters
    ----------
    n : complex
        Mellin moment
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`
    nf : int
        Number of active flavors

    Returns
    -------
    complex
        |NLO| heavy-gluon |OME| :math:`A_{Hg}^{S,(1)}`

    """
    gamma0_qghat = 2 * gamma0_qg(n, nf=1)
    return (1 / 2) * gamma0_qghat * (-L)


@nb.njit(cache=True)
def A_gg(L):
    r"""Compute the |NLO| gluon-gluon |OME| :math:`A_{gg,H}^{S,(1)}`.

    Implements :eqref:`186` of :cite:`Bierenbaum:2022biv`.

    Parameters
    ----------
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    complex
        |NLO| gluon-gluon |OME| :math:`A_{gg,H}^{S,(1)}`

    """
    return A_gg_unpol(L)


@nb.njit(cache=True)
def A_singlet(n, L):
    r"""Compute the |NLO| singlet |OME|.

    .. math::
        A^{S,(1)} = \left(\begin{array}{cc}
        A_{gg,H}^{S,(1)} & 0 & 0\\
        0 & 0 & 0 \\
        A_{hg}^{S,(1)} & 0 & 0
        \end{array}\right)

    Parameters
    ----------
    n : complex
        Mellin moment
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    numpy.ndarray
        |NLO| singlet |OME| :math:`A^{S,(1)}`

    """
    A_S = np.array(
        [
            [A_gg(L), 0.0 + 0j, 0.0 + 0j],
            [0 + 0j, 0 + 0j, 0 + 0j],
            [A_hg(n, L), 0.0 + 0j, 0.0 + 0j],
        ],
        np.complex_,
    )
    return A_S



"""Common type definitions, only used for static analysis."""
import enum
import typing
from typing import Any, Dict, Generic, Tuple, TypeVar

# Energy scales
# -------------

Scalar = float
Scale = float

LinearScale = Scale
SquaredScale = Scale
# TODO: replace with (requires py>=3.9)
#  LinearScale = Annotated[Scale, 1]
#  SquaredScale = Annotated[Scale, 2]


# Flavors
# -------

Order = Tuple[int, int]
FlavorsNumber = int
FlavorIndex = int
IntrinsicFlavors = typing.List[FlavorIndex]
N3LOAdVariation = typing.Tuple[int, int, int, int]

# Evolution coordinates
# ---------------------

EvolutionPoint = Tuple[Scale, FlavorsNumber]


# Scale functions
# ---------------

T = TypeVar("T")


class ReferenceRunning(list, Generic[T]):
    """Running quantities reference point.

    To simplify serialization, the class is just storing the content as a list,
    but:

    - it is constructed with a ``Running.typed(T, Scale)`` signature
    - it should always be used through the property accessors, rather then
      using the list itself

    """

    @classmethod
    def typed(cls, value: T, scale: Scale):
        """Define constructor from individual values.

        This is the preferred constructor for references, since respects the
        intended types of the values.
        It is not the default one only to simplify (de)serialization.

        """
        return cls([value, scale])

    @property
    def value(self) -> T:
        """Reference value, given at a specified scale."""
        return self[0]

    @value.setter
    def value(self, value: T):
        self[0] = value

    @property
    def scale(self) -> Scale:
        """Reference scale, at which the value of the function is given."""
        return self[1]

    @scale.setter
    def scale(self, value: Scale):
        self[1] = value


FlavNumRef = ReferenceRunning[FlavorsNumber]
LinearScaleRef = ReferenceRunning[LinearScale]


# Numerical methods
# -----------------

# TODO: upgrade all the following to StrEnum (requires py>=3.11)
# with that, it is possible to replace all non-alias right sides with calls to
# enum.auto()


class EvolutionMethod(enum.Enum):
    """DGLAP solution method."""

    ITERATE_EXACT = "iterate-exact"
    ITERATE_EXPANDED = "iterate-expanded"
    PERTURBATIVE_EXACT = "perturbative-exact"
    PERTURBATIVE_EXPANDED = "perturbative-expanded"
    TRUNCATED = "truncated"
    ORDERED_TRUNCATED = "ordered-truncated"
    DECOMPOSE_EXACT = "decompose-exact"
    DECOMPOSE_EXPANDED = "decompose-expanded"


class ScaleVariationsMethod(enum.Enum):
    """Method used to account for factorization scale variation."""

    EXPONENTIATED = "exponentiated"
    EXPANDED = "expanded"


class InversionMethod(enum.Enum):
    """Method used to invert the perturbative matching conditions."""

    EXACT = "exact"
    EXPANDED = "expanded"


RawCard = Dict[str, Any]



"""Recipes containing instructions for atomic computations."""
from typing import List

from ..io.items import Evolution, Matching, Recipe
from ..io.types import EvolutionPoint as EPoint
from ..matchings import Atlas, Segment


def elements(ep: EPoint, atlas: Atlas) -> List[Recipe]:
    """Determine recipes to compute a given operator."""
    recipes = []

    #  expanded = eko.operator_card.configs.scvar_method is sv.Modes.expanded
    #  mu2f = mu2 * eko.theory_card.xif**2 if expanded else mu2

    blocks = atlas.matched_path(ep)
    for block in blocks:
        if isinstance(block, Segment):
            cliff = block.target in atlas.walls
            recipe = Evolution.from_atlas(block, cliff=cliff)
        else:
            recipe = Matching.from_atlas(block)

        recipes.append(recipe)

    return recipes


def create(evolgrid: List[EPoint], atlas: Atlas) -> List[Recipe]:
    """Create all associated recipes."""
    recipes = []
    for ep in evolgrid:
        recipes.extend(elements(ep, atlas))

    return list(set(recipes))



"""Define `eko.EKO` metadata."""
import logging
import os
import pathlib
from dataclasses import dataclass
from typing import Optional

import yaml

from .. import version as vmod
from .bases import Bases
from .dictlike import DictLike
from .paths import InternalPaths
from .types import EvolutionPoint as EPoint

logger = logging.getLogger(__name__)


@dataclass
class Metadata(DictLike):
    """Manage metadata, and keep them synced on disk.

    It is possible to have a metadata view, in which the path is not actually
    connected (i.e. it is set to ``None``). In this case, no update will be
    possible, of course.

    Note
    ----
    Unfortunately, for nested structures it is not possible to detect a change
    in their attributes, so a call to :meth:`update` has to be performed
    manually.

    """

    origin: EPoint
    """Inital scale."""
    bases: Bases
    """Manipulation information, describing the current status of the EKO (e.g.
    `inputgrid` and `targetgrid`).
    """
    # tagging information
    _path: Optional[pathlib.Path] = None
    """Path to temporary dir."""
    version: str = vmod.__version__
    """Library version used to create the corresponding file."""
    data_version: int = vmod.__data_version__
    """Specs version, to which the file adheres."""

    @classmethod
    def load(cls, path: os.PathLike):
        """Load metadata from open folder.

        Parameters
        ----------
        path: os.PathLike
            the path to the open EKO folder

        Returns
        -------
        bool
            loaded metadata

        """
        path = pathlib.Path(path)
        content = cls.from_dict(
            yaml.safe_load(InternalPaths(path).metadata.read_text(encoding="utf-8"))
        )
        content._path = path
        return content

    def update(self):
        """Update the disk copy of metadata."""
        if self._path is None:
            logger.info("Impossible to set metadata, no file attached.")
        else:
            with open(InternalPaths(self._path).metadata, "w") as fd:
                yaml.safe_dump(self.raw, fd)

    @property
    def path(self):
        """Access temporary dir path.

        Raises
        ------
        RuntimeError
            if path has not been initialized before

        """
        if self._path is None:
            raise RuntimeError(
                "Access to EKO directory attempted, but not dir has been set."
            )
        return self._path

    @path.setter
    def path(self, value: pathlib.Path):
        """Set temporary dir path."""
        self._path = value

    @property
    def raw(self):
        """Override default :meth:`DictLike.raw` representation to exclude path."""
        return self.public_raw



"""The unpolarized, time-like |NLO| matching conditions."""

import numba as nb
import numpy as np

from eko.constants import CF


@nb.njit(cache=True)
def A_hg(N, L):
    r"""Compute the |NLO| heavy-gluon |OME|.

    Implements :eqref:`27` from :cite:`Cacciari:2005ry`.

    Parameters
    ----------
    N : complex
        Mellin moment
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    A_hg : complex
        |NLO| heavy-gluon |OME|
        :math:`A_{hg}^{S,(1)}`

    """
    result = (
        2
        * CF
        * (
            (2 + N + N**2) / (N * (N**2 - 1)) * (L - 1)
            + 4 / (N - 1) ** 2
            - 4 / N**2
            + 2 / (N + 1) ** 2
        )
    )
    return result


@nb.njit(cache=True)
def A_gg(L):
    r"""Compute the |NLO| gluon-gluon |OME|.

    Implements the Mellin transform of
    :eqref:`24` from :cite:`Cacciari:2005ry`.
    It is identical to the one in
    :class:`~ekore.operator_matrix_elements.unpolarized.space_like.as1`.

    Parameters
    ----------
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    A_gg : complex
        |NLO| gluon-gluon |OME| :math:`A_{gg,H}^{S,(1)}`

    """
    return -2.0 / 3.0 * L


@nb.njit(cache=True)
def A_singlet(N, L):
    r"""Compute the |NLO| singlet |OME|.

    Parameters
    ----------
    N : complex
        Mellin moment
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    numpy.ndarray
        |NLO| singlet |OME|
        :math:`A^{S,(1)}`

    """
    result = np.array(
        [[A_gg(L), 0 + 0j, 0], [0 + 0j, 0, 0], [A_hg(N, L), 0, 0]], np.complex_
    )
    return result


@nb.njit(cache=True)
def A_ns():
    r"""Compute the |NLO| non-singlet |OME|.

    Returns
    -------
    A_ns : numpy.ndarray
        |NLO| non-singlet |OME|
        :math:`A^{S,(1)}`

    """
    result = np.array([[0, 0], [0, 0]], np.complex_)
    return result



"""Create fake PDF sets for debugging."""
import copy
import logging
import pathlib
import shutil
from typing import Callable, List, Optional, Union

import numpy as np

from eko import basis_rotation as br
from eko.io.types import EvolutionPoint as EPoint

from . import export, flavors, load

logger = logging.getLogger(__name__)


def take_data(
    parent_pdf_set: Optional[Union[str, dict]] = None,
    members: bool = False,
    xgrid: Optional[List[float]] = None,
    evolgrid: Optional[List[EPoint]] = None,
):
    """
    Auxiliary function for `generate_pdf`.

    It provides the info, the heads of the member files and the blocks
    to be generated to `generate_pdf`.

    Returns
    -------
        info : dict
            info dictionary
        heads : list(str)
            heads of member files if necessary
        blocks : list(dict)
            data blocks
    """
    if xgrid is None:
        xgrid = np.geomspace(1e-9, 1, 240)
    if evolgrid is None:
        evolgrid = [(mu2, 0) for mu2 in np.geomspace(1.3, 1e5, 35)]
    # collect blocks
    all_blocks = []
    info = None
    heads = []
    if parent_pdf_set is None:
        parent_pdf_set = {
            pid: lambda x, _Q2: x * (1 - x) for pid in br.flavor_basis_pids
        }
    if isinstance(parent_pdf_set, str):
        if parent_pdf_set in ["toy_pol", "toy"]:
            # delay import to ease nnpdf integration
            from banana import toy  # pylint: disable=import-outside-toplevel

            info = copy.deepcopy(load.Toy_info)
            if "pol" in parent_pdf_set:
                toylh = toy.mkPDF("ToyLH_polarized", 0)
            else:
                toylh = toy.mkPDF("", 0)
            all_blocks.append(
                [generate_block(toylh.xfxQ2, xgrid, evolgrid, br.flavor_basis_pids)]
            )

        else:
            info = load.load_info_from_file(parent_pdf_set)
            # iterate on members
            for m in range(int(info["NumMembers"])):
                head, blocks = load.load_blocks_from_file(parent_pdf_set, m)
                heads.append(head)
                all_blocks.append(blocks)
                if not members:
                    break
    elif isinstance(parent_pdf_set, dict):
        info = copy.deepcopy(load.template_info)
        all_blocks.append(
            [
                generate_block(
                    lambda pid, x, Q2: 0.0
                    if pid not in parent_pdf_set
                    else parent_pdf_set[pid](x, Q2),
                    xgrid,
                    evolgrid,
                    br.flavor_basis_pids,
                )
            ]
        )
    else:
        raise ValueError("Unknown parent pdf type")
    return info, heads, all_blocks


def generate_pdf(
    name: str,
    labels: List[int],
    parent_pdf_set=None,
    members=False,
    info_update=None,
    install: bool = False,
    xgrid: Optional[List[float]] = None,
    evolgrid: Optional[List[EPoint]] = None,
):
    """
    Generate a new PDF from a parent PDF with a set of flavors.

    If `parent_pdf_set` is the name of an available PDF set,
    it will be used as parent. In order to use the toy PDF
    as parent, it is enough to set `parent_pdf_set` to "toy".
    In order to use the toy Polarized PDF
    as parent, it is enough to set `parent_pdf_set` to "toy_pol".
    If `parent_pdf_set` is not specified, a debug PDF constructed as
    x * (1-x) for every flavor will be used as parent.
    It is also possible to provide custom functions for each flavor
    in the form of a dictionary: `{pid: f(x,Q2)}`.

    With `labels` it is possible to pass a list of PIDs or evolution basis
    combinations to keep in the generated PDF. In order to project
    on custom combinations of PIDs, it is also possible to pass a list
    containing the desired factors for each flavor.

    The default behavior is to generate only one member for a PDF set
    (the zero member) but it can be changed setting to True the `members` flag.

    The `info_update` argument is a dictionary and provide to the user as a way
    to change the info file of the generated PDF set. If a key of `info_update`
    matches with one key of the standard info file, the information
    are updated, otherwise they are simply added.

    Turning True the value of the `install` flag, it is possible to automatically
    install the generated PDF to the lhapdf directory. By default install is False.

    Examples
    --------
    To generate a PDF with a fixed function `f(x,Q2)` for some flavors
    you can use the following snippet:

    >>> # f = lambda x,Q2 ... put the desired function here
    >>> # mask = [list of active PIDs]
    >>> generate_pdf(name, labels, parent_pdf_set={pid: f for pid in mask})

    The |API| also provides the possibility to extract arbitrary flavor combinations:
    using the debug PDF settings we can construct a "anti-QED-singlet" combination that
    is usefull in debugging DIS codes since it does not couple in |LO|, but only
    through the pure-singlet contributions (starting at |NNLO|)

    >>> from eko import basis_rotation as br
    >>> from ekobox import genpdf
    >>> import numpy as np
    >>> anti_qed_singlet = np.zeros_like(br.flavor_basis_pids, dtype=np.float_)
    >>> anti_qed_singlet[br.flavor_basis_pids.index(1)] = -4
    >>> anti_qed_singlet[br.flavor_basis_pids.index(-1)] = -4
    >>> anti_qed_singlet[br.flavor_basis_pids.index(2)] = 1
    >>> anti_qed_singlet[br.flavor_basis_pids.index(-2)] = 1
    >>> genpdf.generate_pdf("anti_qed_singlet", [anti_qed_singlet])

    """
    pathlib.Path(name).mkdir(exist_ok=True)
    # Checking label basis
    is_evol = False
    flavor_combinations = labels
    if flavors.is_evolution_labels(labels):
        is_evol = True
        flavor_combinations = flavors.evol_to_flavor(labels)
    elif flavors.is_pid_labels(labels):
        labels = np.array(labels, dtype=np.int_)
        flavor_combinations = flavors.pid_to_flavor(labels)

    # labels = verify_labels(args.labels)
    info, heads, all_blocks = take_data(
        parent_pdf_set, members, xgrid=xgrid, evolgrid=evolgrid
    )

    # filter the PDF
    new_all_blocks = []
    for b in all_blocks:
        new_all_blocks.append(flavors.project(b, flavor_combinations))

    # changing info file according to user choice
    if info_update is not None:
        info.update(info_update)
    # write
    info["Flavors"] = [int(pid) for pid in br.flavor_basis_pids]
    info["NumFlavors"] = len(br.flavor_basis_pids)
    if is_evol:
        info["ForcePositive"] = 0
    info["NumMembers"] = len(new_all_blocks)
    # exporting
    export.dump_set(name, info, new_all_blocks, pdf_type_list=heads)

    # install
    if install:
        install_pdf(name)


def install_pdf(name):
    """Install set into LHAPDF.

    The set to be installed has to be in the current directory.

    Parameters
    ----------
    name : str
        source pdf name

    """
    import lhapdf  # pylint: disable=import-error, import-outside-toplevel

    print(f"install_pdf {name}")
    target = pathlib.Path(lhapdf.paths()[0])
    src = pathlib.Path(name)
    if (target / src.stem).exists():
        logger.warning("Overwriting old PDF installation")
        shutil.rmtree(target / src.stem)
    # shutil.move only accepts paths since 3.9 so we need to cast
    # https://docs.python.org/3/library/shutil.html?highlight=shutil#shutil.move
    shutil.move(str(src), str(target))


def generate_block(
    xfxQ2: Callable, xgrid: List[float], evolgrid: List[EPoint], pids: List[int]
) -> dict:
    """Generate an LHAPDF data block from a callable."""
    block: dict = dict(mu2grid=[mu2 for mu2, _ in evolgrid], pids=pids, xgrid=xgrid)
    data = []
    for x in xgrid:
        for mu2, _ in evolgrid:
            data.append(np.array([xfxQ2(pid, x, mu2) for pid in pids]))
    block["data"] = np.array(data)
    return block



"""The |LO| Altarelli-Parisi splitting kernels."""

import numba as nb
import numpy as np

from eko import constants

from ....harmonics import cache as c
from ...unpolarized.space_like.as1 import gamma_ns


@nb.njit(cache=True)
def gamma_qg(N, nf):
    r"""Compute the |LO| polarized quark-gluon anomalous dimension.

    Implements :eqref:`A.1` from :cite:`Gluck:1995yr`.

    Parameters
    ----------
    N : complex
      Mellin moment
    nf : int
      Number of active flavors

    Returns
    -------
    complex
      |LO| polarized quark-gluon anomalous dimension :math:`\gamma_{qg}^{(0)}(N)`

    """
    gamma = -(N - 1) / N / (N + 1)
    result = 2.0 * constants.TR * 2.0 * nf * gamma
    return result


@nb.njit(cache=True)
def gamma_gq(N):
    r"""Compute the |LO| polarized gluon-quark anomalous dimension.

    Implements :eqref:`A.1` from :cite:`Gluck:1995yr`.

    Parameters
    ----------
    N : complex
      Mellin moment

    Returns
    -------
    complex
      |LO| gluon-quark anomalous dimension :math:`\gamma_{gq}^{(0)}(N)`

    """
    gamma = -(N + 2) / N / (N + 1)
    result = 2.0 * constants.CF * gamma
    return result


@nb.njit(cache=True)
def gamma_gg(N, cache, nf):
    r"""Compute the |LO| polarized gluon-gluon anomalous dimension.

    Implements :eqref:`A.1` from :cite:`Gluck:1995yr`.

    Parameters
    ----------
    N : complex
      Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    nf : int
      Number of active flavors

    Returns
    -------
    complex
      |LO| gluon-gluon anomalous dimension :math:`\gamma_{gg}^{(0)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    gamma = -S1 + 2 / N / (N + 1)
    result = constants.CA * (-4.0 * gamma - 11.0 / 3.0) + 4.0 / 3.0 * constants.TR * nf
    return result


@nb.njit(cache=True)
def gamma_singlet(N, cache, nf):
    r"""Compute the |LO| polarized singlet anomalous dimension matrix.

      .. math::
          \gamma_S^{(0)} = \left(\begin{array}{cc}
            \gamma_{qq}^{(0)} & \gamma_{qg}^{(0)}\\
            \gamma_{gq}^{(0)} & \gamma_{gg}^{(0)}
          \end{array}\right)

    Parameters
    ----------
    N : complex
      Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    nf : int
      Number of active flavors

    Returns
    -------
    numpy.ndarray
      |LO| singlet anomalous dimension matrix :math:`\gamma_{S}^{(0)}(N)`

    """
    gamma_qq = gamma_ns(N, cache)
    gamma_S_0 = np.array(
        [[gamma_qq, gamma_qg(N, nf)], [gamma_gq(N), gamma_gg(N, cache, nf)]],
        np.complex_,
    )
    return gamma_S_0



"""Operators bases."""
from dataclasses import dataclass, fields
from typing import Optional

import numpy as np
import numpy.typing as npt

from .. import basis_rotation as br
from .. import interpolation
from .dictlike import DictLike


@dataclass
class Bases(DictLike):
    """Rotations related configurations.

    Here "Rotation" is intended in a broad sense: it includes both rotations in
    flavor space (labeled with suffix `pids`) and in :math:`x`-space (labeled
    with suffix `grid`).
    Rotations in :math:`x`-space correspond to reinterpolate the result on a
    different basis of polynomials.

    """

    xgrid: interpolation.XGrid
    """Internal momentum fraction grid."""
    _targetgrid: Optional[interpolation.XGrid] = None
    _inputgrid: Optional[interpolation.XGrid] = None
    _targetpids: Optional[npt.NDArray] = None
    _inputpids: Optional[npt.NDArray] = None

    def __post_init__(self):
        """Adjust types when loaded from serialized object."""
        for attr in ("xgrid", "_inputgrid", "_targetgrid"):
            value = getattr(self, attr)
            if value is None:
                continue
            if isinstance(value, (np.ndarray, list)):
                setattr(self, attr, interpolation.XGrid(value))
            elif not isinstance(value, interpolation.XGrid):
                setattr(self, attr, interpolation.XGrid.load(value))

    @property
    def pids(self):
        """Internal flavor basis, used for computation."""
        return np.array(br.flavor_basis_pids)

    @property
    def inputpids(self) -> npt.NDArray:
        """Provide pids expected on the input PDF."""
        if self._inputpids is None:
            return self.pids
        return self._inputpids

    @inputpids.setter
    def inputpids(self, value):
        self._inputpids = value

    @property
    def targetpids(self) -> npt.NDArray:
        """Provide pids corresponding to the output PDF."""
        if self._targetpids is None:
            return self.pids
        return self._targetpids

    @targetpids.setter
    def targetpids(self, value):
        self._targetpids = value

    @property
    def inputgrid(self) -> interpolation.XGrid:
        """Provide :math:`x`-grid expected on the input PDF."""
        if self._inputgrid is None:
            return self.xgrid
        return self._inputgrid

    @inputgrid.setter
    def inputgrid(self, value: interpolation.XGrid):
        self._inputgrid = value

    @property
    def targetgrid(self) -> interpolation.XGrid:
        """Provide :math:`x`-grid corresponding to the output PDF."""
        if self._targetgrid is None:
            return self.xgrid
        return self._targetgrid

    @targetgrid.setter
    def targetgrid(self, value: interpolation.XGrid):
        self._targetgrid = value

    @classmethod
    def from_dict(cls, dictionary: dict):
        """Deserialize rotation.

        Load from full state, but with public names.

        """
        d = dictionary.copy()
        for f in fields(cls):
            if f.name.startswith("_"):
                d[f.name] = d.pop(f.name[1:])
        return cls._from_dict(d)

    @property
    def raw(self):
        """Serialize rotation.

        Pass through interfaces, access internal values but with a public name.

        """
        d = self._raw()
        for key in d.copy():
            if key.startswith("_"):
                d[key[1:]] = d.pop(key)

        return d



"""Combine parts into operators."""
from functools import reduce
from typing import List

import numpy as np
import numpy.typing as npt

from ..io.inventory import Inventory
from ..io.items import Evolution, Operator, Recipe


def retrieve(
    headers: List[Recipe], parts: Inventory, parts_matching: Inventory
) -> List[Operator]:
    """Retrieve parts to be joined."""
    elements = []
    for head in headers:
        inv = parts if isinstance(head, Evolution) else parts_matching
        elements.append(inv[head])

    return elements


def dot4(op1: npt.NDArray, op2: npt.NDArray) -> npt.NDArray:
    """Dot product between rank 4 objects.

    The product is performed considering them as matrices indexed by pairs, so
    linearizing the indices in pairs.

    """
    return np.einsum("aibj,bjck->aick", op1, op2)


def dotop(op1: Operator, op2: Operator) -> Operator:
    r"""Dot product between two operators.

    Essentially a wrapper of :func:`dot4`, applying linear error propagation,
    if applicable.

    Note
    ----
    Linear error propagation requires matrices to be positive before taking the product.

    Indeed, for a simple product of two variables :math:`a \cdot b`, the error is
    propagated in the following way:

    .. math::

        \max_{\sgn_{da}, \sgn{db}} (a + da)(b + db) - ab =
        \max_{\sgn_{da}, \sgn{db}} da \cdot b + a \cdot db + \mathcal{O}(d^2) =
        |da \cdot b| + |a \cdot db| + \mathcal{O}(d^2) =
        |da | \cdot |b| + |a| \cdot |db| + \mathcal{O}(d^2)

    Where the second step is a consequence of being able to vary the two
    variations independently, and last just a trivial property of the product.

    But in a matrix multiplication, each element of the two matrices has an
    independent variation associated. Thus:

    .. math::

        \max_{\sgn_{da_i}, \sgn{db_i}} (a_i + da_i)(b_i + db_i) - a_i b_i =
        \max_{\sgn_{da_i}, \sgn{db_i}} da_i \cdot b_i + a_i \cdot db_i + \mathcal{O}(d^2) =
        |da_i| \cdot |b_i| + |a_i| \cdot |db_i| + \mathcal{O}(d^2)

    """
    val = dot4(op1.operator, op2.operator)

    if op1.error is not None and op2.error is not None:
        err = dot4(np.abs(op1.operator), np.abs(op2.error)) + dot4(
            np.abs(op1.error), np.abs(op2.operator)
        )
    else:
        err = None

    return Operator(val, err)


def join(elements: List[Operator]) -> Operator:
    """Join the elements into the final operator.

    Note
    ----
    Since the matrices composing the path have to be multiplied from the
    destination to the origin, the input order, coming from path (which is
    instead ``origin -> target``), is being reversed.

    .. todo::

        consider if reversing the path...

    """
    return reduce(dotop, reversed(elements))



"""Operator card configurations."""
from banana.data import cartesian_product, sql

from eko import interpolation

from . import db

default_card = dict(
    sorted(
        dict(
            interpolation_xgrid=interpolation.make_grid(30, 20).tolist(),
            interpolation_polynomial_degree=4,
            interpolation_is_log=True,
            ev_op_max_order=10,
            ev_op_iterations=10,
            backward_inversion="expanded",
            n_integration_cores=0,
            debug_skip_non_singlet=False,
            debug_skip_singlet=False,
            mugrid=[10],
            inputgrid=None,
            targetgrid=None,
            inputpids=None,
            targetpids=None,
            polarized=False,
            time_like=False,
        ).items()
    )
)


lhapdf_config = {
    # "ev_op_max_order": [10],
    # "ev_op_iterations": [2, 10, 30],
    "mugrid": [[4.4, 10, 31, 100]],
}

apfel_config = {
    # "ev_op_max_order": [10],
    # "ev_op_iterations": [2, 10, 30],
    "mugrid": [[31, 100]],
}

pegasus_config = {
    "ev_op_max_order": [10],
    "ev_op_iterations": [10],
    "mugrid": [[31]],
}

pegasus_exact_config = {
    "ev_op_max_order": [15],
    "ev_op_iterations": [20],
    "mugrid": [[31]],
}


def build(update=None):
    """
    Generate all operator card updates.

    Parameters
    ----------
        update : dict
            base modifiers

    Returns
    -------
        cards : list(dict)
            list of update
    """
    cards = []
    if update is None:
        update = {}
    for c in cartesian_product(update):
        card = {}
        card.update(c)
        cards.append(card)
    return cards


# db interface
def load(session, updates):
    """
    Load operator records from the DB.

    Parameters
    ----------
    session : sqlalchemy.session.Session
        DB ORM session
    updates : dict
        modifiers

    Returns
    -------
    cards : list(dict)
        list of records
    """
    # add hash
    raw_records, df = sql.prepare_records(default_card, updates)

    # insert new ones
    sql.insertnew(session, db.Operator, df)
    return raw_records



"""Utility functions."""

import numba as nb
import numpy as np


@nb.njit(cache=True)
def geomspace(start, end, num):
    """Numba port of :func:`numpy.geomspace`.

    Parameters
    ----------
    start : float
        initial value
    end : float
        final value
    num : int
        steps

    Returns
    -------
    geomspace : numpy.ndarray
        logarithmic spaced list between `start` and `end`

    """
    return np.exp(np.linspace(np.log(start), np.log(end), num))



"""Weight 4 harmonic sums."""

import numba as nb

from eko.constants import li4half, log2, zeta2, zeta3, zeta4

from . import g_functions as gf
from .polygamma import cern_polygamma, symmetry_factor


@nb.njit(cache=True)
def S4(N):
    r"""Compute the harmonic sum :math:`S_4(N)`.

    .. math::
      S_4(N) = \sum\limits_{j=1}^N \frac 1 {j^4} = - \frac 1 6 \psi_3(N+1)+\zeta(4)

    with :math:`\psi_3(N)` the 3rd-polygamma function and :math:`\zeta` the
    Riemann zeta function.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    S_4 : complex
        Harmonic sum :math:`S_4(N)`

    See Also
    --------
    ekore.harmonics.polygamma.cern_polygamma : :math:`\psi_k(N)`

    """
    return zeta4 - 1.0 / 6.0 * cern_polygamma(N + 1.0, 3)


@nb.njit(cache=True)
def Sm4(N, hS4, hS4mh, hS4h, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-4}(N)`.

    .. math::
      S_{-4}(N) = \sum\limits_{j=1}^N \frac {(-1)^j} {j^4}

    Parameters
    ----------
    N : complex
        Mellin moment
    hS4:  complex
        Harmonic sum :math:`S_{4}(N)`
    hS4mh: complex
        Harmonic sum :math:`S_{4}((N-1)/2)`
    hS4h: complex
        Harmonic sum :math:`S_{4}(N/2)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N =
        1`), False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm4 : complex
        Harmonic sum :math:`S_{-4}(N)`

    See Also
    --------
    eko.anomalous_dimension.w4.S4 : :math:`S_4(N)`

    """
    if is_singlet is None:
        return (
            1 / 2**3 * ((1 - (-1) ** N) / 2 * hS4mh + ((-1) ** N + 1) / 2 * hS4h)
            - hS4
        )
    if is_singlet:
        return 1 / 2**3 * hS4h - hS4
    return 1 / 2**3 * hS4mh - hS4


@nb.njit(cache=True)
def Sm31(N, S1, Sm1, Sm2, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-3,1}(N)`.

    As implemented in :eqref:`B.5.93` of :cite:`MuselliPhD` and :eqref:`25` of
    cite:`Bl_mlein_2000`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    Sm1: complex
        Harmonic sum :math:`S_{-1}(N)`
    Sm2: complex
        Harmonic sum :math:`S_{-2}(N)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N =
        1`), False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm31 : complex
        Harmonic sum :math:`S_{-3,1}(N)`

    See Also
    --------
    ekore.harmonics.g_functions.mellin_g6 : :math:`g_6(N)`

    """
    eta = symmetry_factor(N, is_singlet)
    return (
        eta * gf.mellin_g6(N, S1)
        + zeta2 * Sm2
        - zeta3 * Sm1
        - 3 / 5 * zeta2**2
        + 2 * li4half
        + 3 / 4 * zeta3 * log2
        - 1 / 2 * zeta2 * log2**2
        + 1 / 12 * log2**4
    )


@nb.njit(cache=True)
def Sm22(N, S1, S2, Sm2, Sm31, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-2,2}(N)`.

    As implemented in :eqref:`B.5.94` of :cite:`MuselliPhD` and :eqref:`24` of
    cite:`Bl_mlein_2000`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    Sm2: complex
        Harmonic sum :math:`S_{-2}(N)`
    Sm31: complex
        Harmonic sum :math:`S_{-3,1}(N)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N =
        1`), False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm22 : complex
        Harmonic sum :math:`S_{-2,2}(N)`

    See Also
    --------
    ekore.harmonics.g_functions.mellin_g5 : :math:`g_5(N)`

    """
    eta = symmetry_factor(N, is_singlet)
    return (
        eta * gf.mellin_g5(N, S1, S2) - 2 * Sm31 + 2 * zeta2 * Sm2 + 3 / 40 * zeta2**2
    )


@nb.njit(cache=True)
def Sm211(N, S1, S2, Sm1, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-2,1,1}(N)`.

    As implemented in :eqref:`B.5.104` of :cite:`MuselliPhD` and :eqref:`27` of
    cite:`Bl_mlein_2000`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    Sm1: complex
        Harmonic sum :math:`S_{-1}(N)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N =
        1`), False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm221 : complex
        Harmonic sum :math:`S_{-2,1,1}(N)`

    See Also
    --------
    ekore.harmonics.g_functions.mellin_g8 : :math:`g_8(N)`

    """
    eta = symmetry_factor(N, is_singlet)
    return (
        -eta * gf.mellin_g8(N, S1, S2)
        + zeta3 * Sm1
        - li4half
        + 1 / 8 * zeta2**2
        + 1 / 8 * zeta3 * log2
        + 1 / 4 * zeta2 * log2**2
        - 1 / 24 * log2**4
    )


@nb.njit(cache=True)
def S211(N, S1, S2, S3):
    r"""Analytic continuation of harmonic sum :math:`S_{2,1,1}(N)`.

    As implemented in :eqref:`B.5.115` of :cite:`MuselliPhD` and :eqref:`40` of
    cite:`Bl_mlein_2000`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    S3: complex
        Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
    S211 : complex
        Harmonic sum :math:`S_{2,1,1}(N)`

    See Also
    --------
    ekore.harmonics.g_functions.mellin_g21 : :math:`g_21(N)`

    """
    return -gf.mellin_g21(N, S1, S2, S3) + 6 / 5 * zeta2**2


@nb.njit(cache=True)
def S31(N, S1, S2, S3, S4):
    r"""Analytic continuation of harmonic sum :math:`S_{3,1}(N)`.

    As implemented in :eqref:`B.5.99` of :cite:`MuselliPhD` and :eqref:`41` of
    cite:`Bl_mlein_2000`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    S3: complex
        Harmonic sum :math:`S_{3}(N)`
    S4: complex
        Harmonic sum :math:`S_{4}(N)`

    Returns
    -------
    S31 : complex
        Harmonic sum :math:`S_{3,1}(N)`

    See Also
    --------
    ekore.harmonics.g_functions.mellin_g22 : :math:`g_22(N)`

    """
    return (
        1 / 2 * gf.mellin_g22(N, S1, S2, S3)
        - 1 / 4 * S4
        - 1 / 4 * S2**2
        + zeta2 * S2
        - 3 / 20 * zeta2**2
    )



"""Library of reusable options and elements."""
import pathlib

import click

THEORY = "theory.yaml"
OPERATOR = "operator.yaml"
OUTPUT = "eko.tar"


def destination(dest: pathlib.Path):
    """Build click option for destination.

    Parameters
    ----------
    dest : pathlib.Path
        default destination

    Returns
    -------
    click.option
        generated option

    """
    return click.option(
        "-d",
        "--destination",
        type=click.Path(path_type=pathlib.Path, exists=True),
        default=dest,
        help="Alternative destination path to store the resulting table (default: $PWD/theory)",
    )



r"""Tools for Mellin inversion.

We provide all necessary toold to deal with the
`inverse Mellin transformation <https://en.wikipedia.org/wiki/Mellin_inversion_theorem>`_.

Although this module provides three different path implementations in practice
only the Talbot path :cite:`Abate`

.. math::
    p_{\text{Talbot}}(t) =  o + r \cdot ( \theta \cot(\theta) + i\theta)\quad
        \text{with}~\theta = \pi(2t-1)

is used, as it results in the most efficient convergence. The default values
for the parameters :math:`r,o` are given by :math:`r = 1/2, o = 0` for
the non-singlet integrals and by :math:`r = \frac{2}{5} \frac{16}{1 - \ln(x)}, o = 1`
for the singlet sector. Note that the non-singlet kernels evolve poles only up to
:math:`N=0` whereas the singlet kernels have poles up to :math:`N=1`.

"""

import numba as nb
import numpy as np


@nb.njit(cache=True)
def Talbot_path(t, r, o):
    r"""Compute Talbot path.

    .. math::
        p_{\text{Talbot}}(t) =  o + r \cdot ( \theta \cot(\theta) + i\theta ),
        \theta = \pi(2t-1)

    Parameters
    ----------
    t : float
        way parameter
    r : float
        scaling parameter - effectively corresponds to the intersection of the path with the
        real axis
    o : float
        offset on real axis

    Returns
    -------
    path : complex
        Talbot path

    """
    theta = np.pi * (2.0 * t - 1.0)
    re = 0.0
    if t == 0.5:  # treat singular point separately
        re = 1.0
    else:
        re = theta / np.tan(theta)
    im = theta
    return o + r * complex(re, im)


@nb.njit(cache=True)
def Talbot_jac(t, r, _o):
    r"""Compute Derivative of Talbot path.

    .. math::
        \frac{dp_{\text{Talbot}}(t)}{dt}

    Parameters
    ----------
    t : float
        way parameter
    r : float
        scaling parameter - effectively corresponds to the intersection of the path with the
        real axis
    o : float
        offset on real axis

    Returns
    -------
    jac : complex
        derivative of Talbot path

    """
    theta = np.pi * (2.0 * t - 1.0)
    re = 0.0
    if t == 0.5:  # treat singular point separately
        re = 0.0
    else:
        re = 1.0 / np.tan(theta)
        re -= theta / (np.sin(theta)) ** 2
    im = 1.0
    return r * np.pi * 2.0 * complex(re, im)


@nb.njit(cache=True)
def line_path(t, m, c):
    r"""Compute textbook path, i.e. a straight line parallel to the imaginary axis.

    .. math::
        p_{\text{line}}(t) = c + m \cdot (2t - 1)

    Parameters
    ----------
    t : float
        way parameter
    m : float
        scaling parameter
    c : float
        offset on real axis

    Returns
    -------
    path : complex
        Textbook path

    """
    return complex(c, m * (2 * t - 1))


@nb.njit(cache=True)
def line_jac(_t, m, _c):
    r"""Compute derivative of Textbook path.

    .. math::
        \frac{dp_{\text{line}}(t)}{dt}

    Parameters
    ----------
    t : float
        way parameter
    m : float
        scaling parameter
    o : float
        offset on real axis

    Returns
    -------
    jac : complex
        derivative of Textbook path

    """
    return complex(0, m * 2)


@nb.njit(cache=True)
def edge_path(t, m, c, phi):
    r"""Compute edged path with a given angle.

    .. math::
        p_{\text{edge}}(t) = c + m\left|t - \frac 1 2\right|\exp(i\phi)

    Parameters
    ----------
    t : float
        way parameter
    m : float
        length of the path
    c : float, optional
        intersection of path with real axis
    phi : complex, optional
        bended angle

    Returns
    -------
    path : complex
        Edged path

    """
    if t < 0.5:  # turning point: path is not differentiable in this point
        return c + (0.5 - t) * m * np.exp(complex(0, -phi))
    return c + (t - 0.5) * m * np.exp(complex(0, +phi))


@nb.njit(cache=True)
def edge_jac(t, m, _c, phi):
    r"""Compute derivative of edged path.

    .. math::
        \frac{dp_{\text{edge}}(t)}{dt}

    Parameters
    ----------
    t : float
        way parameter
    m : float
        length of the path
    c : float, optional
        intersection of path with real axis
    phi : complex, optional
        bended angle

    Returns
    -------
    path : complex
        Derivative of edged path

    """
    if t < 0.5:  # turning point: jacobian is not continuous here
        return -m * np.exp(complex(0, -phi))
    return +m * np.exp(complex(0, phi))


spec = [
    ("t", nb.float64),
    ("r", nb.float64),
    ("o", nb.int8),
]


@nb.experimental.jitclass(spec)
class Path:
    """Mellin path dispatcher.

    Parameters
    ----------
    t : float
        way parameter
    logx : float
        Mellin inversion point
    axis_offset: bool
        add offset on the real axis

    """

    def __init__(self, t, logx, axis_offset):
        self.t = t
        # TODO: shall we use: 0.4 * 16.0 / ( - logx) ?
        self.r = 0.4 * 16.0 / (1.0 - logx)
        if axis_offset:
            self.o = 1.0
        else:
            self.o = 0.0

    # TODO: make also the other 2 paths available ??
    @property
    def n(self):
        """Return the Mellin moment N."""
        return Talbot_path(self.t, self.r, self.o)

    @property
    def jac(self):
        """Return the Jacobian of the Mellin path."""
        return Talbot_jac(self.t, self.r, self.o)

    @property
    def prefactor(self):
        r"""Return the mellin inversion prefactor :math:`-\frac{i}{\pi}`."""
        return complex(0.0, -1.0 / np.pi)
